# auto generated. yay.
import ooxcb
from ooxcb.resource import get_internal
from ooxcb.types import SIZES, make_array, build_list
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO
from struct import pack, unpack, calcsize
from ooxcb.protocol.xproto import Window
from ooxcb.util import Mixin

def unpack_from_stream(fmt, stream, offset=0):
    if offset:
        stream.seek(offset, 1)
    s = stream.read(calcsize(fmt))
    return unpack(fmt, s)


MAJOR_VERSION = 1
MINOR_VERSION = 1
key = ooxcb.ExtensionKey("SHAPE")

class SK(object):
    Bounding = 0
    Clip = 1
    Input = 2

class SO(object):
    Set = 0
    Union = 1
    Intersect = 2
    Subtract = 3
    Invert = 4

class shapeExtension(ooxcb.Extension):
    header = "shape"
    def query_version(self):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxx", ))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, True), \
            QueryVersionCookie(),
            QueryVersionReply)

    def query_version_unchecked(self):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxx", ))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, False), \
            QueryVersionCookie(),
            QueryVersionReply)

class WindowMixin(Mixin):
    target_class = Window
    def shape_rectangles_checked(self, operation, destination_kind, ordering, x_offset, y_offset, rectangles):
        rectangles_len = len(rectangles)
        destination_window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBBBxIhh", operation, destination_kind, ordering, destination_window, x_offset, y_offset))
        rectangles.build(buf)
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, True, True), \
            ooxcb.VoidCookie())

    def shape_rectangles(self, operation, destination_kind, ordering, x_offset, y_offset, rectangles):
        rectangles_len = len(rectangles)
        destination_window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBBBxIhh", operation, destination_kind, ordering, destination_window, x_offset, y_offset))
        rectangles.build(buf)
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, True, False), \
            ooxcb.VoidCookie())

    def shape_mask_checked(self, operation, destination_kind, x_offset, y_offset, source_bitmap):
        destination_window = get_internal(self)
        source_bitmap = get_internal(source_bitmap)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBBxxIhhI", operation, destination_kind, destination_window, x_offset, y_offset, source_bitmap))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, True), \
            ooxcb.VoidCookie())

    def shape_mask(self, operation, destination_kind, x_offset, y_offset, source_bitmap):
        destination_window = get_internal(self)
        source_bitmap = get_internal(source_bitmap)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBBxxIhhI", operation, destination_kind, destination_window, x_offset, y_offset, source_bitmap))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, False), \
            ooxcb.VoidCookie())

    def shape_combine_checked(self, operation, destination_kind, source_kind, x_offset, y_offset, source_window):
        destination_window = get_internal(self)
        source_window = get_internal(source_window)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBBBxIhhI", operation, destination_kind, source_kind, destination_window, x_offset, y_offset, source_window))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, True), \
            ooxcb.VoidCookie())

    def shape_combine(self, operation, destination_kind, source_kind, x_offset, y_offset, source_window):
        destination_window = get_internal(self)
        source_window = get_internal(source_window)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBBBxIhhI", operation, destination_kind, source_kind, destination_window, x_offset, y_offset, source_window))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, False), \
            ooxcb.VoidCookie())

    def shape_offset_checked(self, destination_kind, x_offset, y_offset):
        destination_window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBxxxIhh", destination_kind, destination_window, x_offset, y_offset))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 4, True, True), \
            ooxcb.VoidCookie())

    def shape_offset(self, destination_kind, x_offset, y_offset):
        destination_window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBxxxIhh", destination_kind, destination_window, x_offset, y_offset))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 4, True, False), \
            ooxcb.VoidCookie())

    def shape_query_extents(self):
        destination_window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", destination_window))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 5, False, True), \
            QueryExtentsCookie(),
            QueryExtentsReply)

    def shape_query_extents_unchecked(self):
        destination_window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", destination_window))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 5, False, False), \
            QueryExtentsCookie(),
            QueryExtentsReply)

    def shape_select_input_checked(self, enable):
        destination_window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", destination_window, enable))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 6, True, True), \
            ooxcb.VoidCookie())

    def shape_select_input(self, enable):
        destination_window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", destination_window, enable))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 6, True, False), \
            ooxcb.VoidCookie())

    def shape_input_selected(self):
        destination_window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", destination_window))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 7, False, True), \
            InputSelectedCookie(),
            InputSelectedReply)

    def shape_input_selected_unchecked(self):
        destination_window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", destination_window))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 7, False, False), \
            InputSelectedCookie(),
            InputSelectedReply)

    def shape_get_rectangles(self, source_kind):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", window, source_kind))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 8, False, True), \
            GetRectanglesCookie(),
            GetRectanglesReply)

    def shape_get_rectangles_unchecked(self, source_kind):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", window, source_kind))
        return self.conn.shape.send_request(ooxcb.Request(self.conn, buf.getvalue(), 8, False, False), \
            GetRectanglesCookie(),
            GetRectanglesReply)

class QueryVersionReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.major_version = None
        self.minor_version = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xxxxxxxxHH", stream)
        self.major_version = _unpacked[0]
        self.minor_version = _unpacked[1]

    def build(self, stream):
        count = 0
        stream.write(pack("=xxxxxxxxHH", self.major_version, self.minor_version))

class QueryVersionCookie(ooxcb.Cookie):
    pass

class NotifyEvent(ooxcb.Event):
    event_name = "on_notify"
    opcode = 0
    event_target_class = "Window"
    def __init__(self, conn):
        ooxcb.Event.__init__(self, conn)
        self.response_type = 0
        self.shape_kind = None
        self.affected_window = None
        self.extents_x = None
        self.extents_y = None
        self.extents_width = None
        self.extents_height = None
        self.server_time = None
        self.shaped = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=BBxxIhhHHIBxxxxxxxxxxx", stream)
        self.response_type = _unpacked[0]
        self.shape_kind = _unpacked[1]
        self.affected_window = Window(self.conn, _unpacked[2])
        self.extents_x = _unpacked[3]
        self.extents_y = _unpacked[4]
        self.extents_width = _unpacked[5]
        self.extents_height = _unpacked[6]
        self.server_time = _unpacked[7]
        self.shaped = _unpacked[8]
        self.event_target = self.affected_window

    def build(self, stream):
        count = 0
        root = stream.tell()
        stream.write(pack("=BBxxIhhHHIBxxxxxxxxxxx", self.response_type, self.shape_kind, get_internal(self.affected_window), self.extents_x, self.extents_y, self.extents_width, self.extents_height, self.server_time, self.shaped))
        stream.write("\0" * (32 - (stream.tell() - root)))

class GetRectanglesReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.ordering = None
        self.rectangles_len = None
        self.rectangles = []

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xBxxxxxxI", stream)
        self.ordering = _unpacked[0]
        self.rectangles_len = _unpacked[1]
        self.rectangles = ooxcb.List(self.conn, stream, self.rectangles_len, RECTANGLE, 8)

    def build(self, stream):
        count = 0
        stream.write(pack("=xBxxxxxxI", self.ordering, self.rectangles_len))
        count += 12
        build_list(self.conn, stream, self.rectangles, RECTANGLE)

class QueryExtentsCookie(ooxcb.Cookie):
    pass

class InputSelectedReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.enabled = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xBxxxxxx", stream)
        self.enabled = _unpacked[0]

    def build(self, stream):
        count = 0
        stream.write(pack("=xBxxxxxx", self.enabled))

class QueryExtentsReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.bounding_shaped = None
        self.clip_shaped = None
        self.bounding_shape_extents_x = None
        self.bounding_shape_extents_y = None
        self.bounding_shape_extents_width = None
        self.bounding_shape_extents_height = None
        self.clip_shape_extents_x = None
        self.clip_shape_extents_y = None
        self.clip_shape_extents_width = None
        self.clip_shape_extents_height = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xxxxxxxxBBxxhhHHhhHH", stream)
        self.bounding_shaped = _unpacked[0]
        self.clip_shaped = _unpacked[1]
        self.bounding_shape_extents_x = _unpacked[2]
        self.bounding_shape_extents_y = _unpacked[3]
        self.bounding_shape_extents_width = _unpacked[4]
        self.bounding_shape_extents_height = _unpacked[5]
        self.clip_shape_extents_x = _unpacked[6]
        self.clip_shape_extents_y = _unpacked[7]
        self.clip_shape_extents_width = _unpacked[8]
        self.clip_shape_extents_height = _unpacked[9]

    def build(self, stream):
        count = 0
        stream.write(pack("=xxxxxxxxBBxxhhHHhhHH", self.bounding_shaped, self.clip_shaped, self.bounding_shape_extents_x, self.bounding_shape_extents_y, self.bounding_shape_extents_width, self.bounding_shape_extents_height, self.clip_shape_extents_x, self.clip_shape_extents_y, self.clip_shape_extents_width, self.clip_shape_extents_height))

class GetRectanglesCookie(ooxcb.Cookie):
    pass

class InputSelectedCookie(ooxcb.Cookie):
    pass

_events = {
    0: NotifyEvent,
}

_errors = {
}

for ev in _events.itervalues():
    if isinstance(ev.event_target_class, str):
        ev.event_target_class = globals()[ev.event_target_class]

ooxcb._add_ext(key, shapeExtension, _events, _errors)
def mixin():
    WindowMixin.mixin()





# Copyright (c) 2008-2011, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
    This module contains some helper functions for the icccm standard.
"""

from ooxcb.protocol.xproto import Window, Pixmap

class WMState(object):
    """
        container class for the WM_STATE property.

        .. attribute:: state

            one member of :class:`ooxcb.xproto.WMState`

        .. attribute:: icon

            The icon window, either a :class:`ooxcb.xproto.Window` instance
            or None.
    """
    def __init__(self, state, icon):
        self.state = state
        self.icon = icon

def icccm_get_wm_state(window):
    """
        return a :class:`WMState` instance or None if
        there is no `WM_STATE` property.
    """
    prop = window.get_property('WM_STATE', 'WM_STATE').reply()
    if not prop.exists:
        return None
    else:
        icon = None
        if prop.value[1]:
            prop = window.conn.get_from_cache_fallback(prop.value[1], Window)
        return WMState(prop.value[0], icon)

class WMHints(object):
    """
        a container class for the `WM_HINTS` property.

        .. attribute:: flags

            a mask of the :class:`WMHintsFlags` values

        .. attribute:: input

            The client's input model (True/False)

        .. attribute:: initial_state

            The state when first mapped

        .. attribute:: icon_pixmap

            The pixmap for the icon image

        .. attribute:: icon_window

            The window for the icon image

        .. attribute:: icon_x

            X coordinate of the icon location

        .. attribute:: icon_y

            Y coordinate of the icon location

        .. attribute:: icon_mask

            The mask for the icon shape

    """
    def __init__(self, flags, input, initial_state, icon_pixmap,
            icon_window, icon_x, icon_y, icon_mask):
        self.flags = flags
        self.input = input
        self.initial_state = initial_state
        self.icon_pixmap = icon_pixmap
        self.icon_window = icon_window
        self.icon_x = icon_x
        self.icon_y = icon_y
        self.icon_mask = icon_mask

def icccm_get_wm_hints(window):
    """
        return a :class:`WMHints` instance or None if there is no
        `WM_HINTS` property or the value is invalid.
    """
    prop = window.get_property('WM_HINTS', 'CARDINAL').reply()
    if (not prop.exists or not prop.value):
        return None
    else:
        value = prop.value
        return WMHints(
                value[0],
                value[1],
                value[2],
                window.conn.get_from_cache_fallback(value[3], Pixmap),
                window.conn.get_from_cache_fallback(value[4], Window),
                value[5],
                value[6],
                window.conn.get_from_cache_fallback(value[7], Pixmap)
                )

def mixin():
    from ooxcb.util import mixin_functions
    mixin_functions((
        icccm_get_wm_state,
        icccm_get_wm_hints,
        ), Window)




# Copyright (c) 2008-2011, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ctypes
from contextlib import contextmanager
from weakref import WeakValueDictionary

from . import libxcb
from .protobj import Event, Error
from .eventsys import EventDispatcher
from .resource import XNone
from .atoms import AtomDict
from .keysyms import Keysyms

class Connection(EventDispatcher):
    """
        The ooxcb connection class is the main point of ooxcb. It is
        similar to libX11's `Display`.

        The Connection class is an event dispatcher, so you can register
        for events.

        The Connection instance has a `core` attribute that is the core
        extension instance, and it also sets the `xproto` attribute for
        the xproto extension.

        :todo: describe events
    """
    def __init__(self, core):
        """
            :param core: The core extension class.
        """
        EventDispatcher.__init__(self)

        self.core = core(self)
        setattr(self, core.header, self.core)

        # opaque libxcb connection
        self.conn = None
        self.events = {}
        self.errors = {}
        self.extcache = {}
        self._setup = None
        # check all events synchronously? for debugging.
        self.synchronous_check = False
        # helper
        self.keysyms = Keysyms(self)
        # helper
        self.atoms = AtomDict(self)
        # the X object cache. {X ID: Object, ...}
        self._cache = WeakValueDictionary()
        # add XNone to the cache ...
        self.add_to_cache(0, XNone)

    @property
    def pref_screen_object(self):
        """
            Return the preferred :class:`protocol.xproto.Screen` instance.
            Shortcut for ``conn.setup.roots[conn.pref_screen]``
        """
        return self.setup.roots[self.pref_screen]

    @contextmanager
    def bunch(self):
        """
            Use this in a `with` statement.
            Example:

            ::

                with conn.bunch():
                    for window in windows:
                        window.map()

            When the `with` statement is left,
            the connection will be flushed.

        """
        yield
        self.flush()

    def do_initial_setup(self):
        """
            loads the core events, the core errors and sets them up.

            :note: Internally called by :func:`ooxcb.connect`.
        """
        import ooxcb
        # load core ...
        ext = self.core
        events = ooxcb.CORE_EVENTS
        errors = ooxcb.CORE_ERRORS
        self.setup_helper(ext, events, errors)

        for key, events in ooxcb.EXT_EVENTS.iteritems():
            errors = ooxcb.EXT_ERRORS[key]
            ext = self.load_ext(key)

            if ext.present:
                self.setup_helper(ext, events, errors)

    def setup_helper(self, ext, events, errors):
        """
            gets all events and errors from *ext* and registers
            them.

            :note: Internal method.
        """
        for num, type in events.iteritems():
            opcode = ext.first_event + num
            self.events[opcode] = type
            # register all events to their target!
            type.event_target_class.register_event_type(type.event_name)

        for num, type in errors.iteritems():
            opcode = ext.first_error + num
            self.errors[opcode] = type

    def load_ext(self, key):
        """
            load an extension.

            :note: internally called
        """
        import ooxcb
        if key in self.extcache:
            return self.extcache[key]
        else:
            cls = ooxcb.EXTDICT[key]
            ext = cls(self, key)
            reply = libxcb.xcb_get_extension_data(self.conn, key.key).contents
            ext.present = reply.present
            ext.major_opcode = reply.major_opcode
            ext.first_event = reply.first_event
            ext.first_error = reply.first_error

            self.extcache[key] = ext
            setattr(self, cls.header, ext)

            return ext

    @property
    def alive(self):
        """
            True if the connection is valid, False if it
            is not (ie disconnected)
        """
        return self.conn is not None

    def check_conn(self):
        """
            checks *self.conn* and raises an :class:`IOError` if it's None.
        """
        if not self.alive:
            raise IOError("Connection is not valid anymore")

    def has_error(self):
        """
            returns True if the connection has encountered an error.
        """
        self.check_conn()
        return bool(libxcb.xcb_connection_has_error(self.conn))

    def get_file_descriptor(self):
        """
            get the unix file descriptor.
        """
        self.check_conn()
        return libxcb.xcb_get_file_descriptor(self.conn)

    def get_maximum_request_length(self):
        """
            return the connection's maxmimum request length.
        """
        self.check_conn()
        return libxcb.xcb_connection_get_maximum_request_length(self.conn)

    def prefetch_maximum_request_length(self):
        """
            prefetches the maximum request length without blocking.
        """
        self.check_conn()
        libxcb.xcb_prefetch_maximum_request_length(self.conn)

    def get_setup(self):
        """
            get the connection's setup and return it.
            The setup instance is cached.
        """
        self.check_conn()
        if self._setup is None:
            s = libxcb.xcb_get_setup(self.conn)
            addr = ctypes.addressof(s.contents)
            from . import SETUP
            self._setup = SETUP.create_from_address(self, addr)

        return self._setup

    setup = property(get_setup)

    def generate_id(self):
        """
            generates a new X ID and return it.
        """
        self.check_conn()
        xid = libxcb.xcb_generate_id(self.conn)
        # TODO: error checking ...
        return xid

    def wait_for_event(self):
        """
            waits for an event (blocking) and returns it.

            If an connection IO error encountered, it raises an `IOError`.
            This can also raise Xcb protocol errors.
        """
        data = libxcb.xcb_wait_for_event(self.conn)
        if not data:
            raise IOError("I/O error on X server connection.")
        if data.contents.response_type == 0:
            Error.set(self, ctypes.cast(data,
                ctypes.POINTER(libxcb.xcb_generic_error_t)))
            return
        return Event.create(self, data)

    def poll_for_event(self):
        """
            polls for an event (non-blocking) and returns it.
            If there was no event, None is returned.

            :see: *wait_for_event*
        """
        data = libxcb.xcb_poll_for_event(self.conn)
        if not data:
            return None

        if data.contents.response_type == 0:
            Error.set(self, ctypes.cast(data,
                ctypes.POINTER(libxcb.xcb_generic_error_t)))
            return
        return Event.create(self, data)

    def flush(self):
        """
            flushes the connection. All cached requests are sent.
        """
        self.check_conn()
        libxcb.xcb_flush(self.conn)

    def disconnect(self):
        """
            disconnect from the X server gracefully and set
            *self.conn* to None.
        """
        if self.conn:
            libxcb.xcb_disconnect(self.conn)
            self.conn = None

    def add_to_cache(self, xid, obj):
        """
            add the X object *obj* with the X ID *xid*
            to the internal X object cache.
        """
        self._cache[xid] = obj

    def get_from_cache_fallback(self, xid, cls):
        """
            If there is a resource using the xid *xid* in the cache,
            return it. If not, instantiate *cls*, add to cache
            and return the newly created object.
        """
        if xid in self._cache:
            return self._cache[xid]
        else:
            self._cache[xid] = ret = cls(self, xid)
            return ret

    def remove_from_cache(self, xid):
        """
            removes the object managing the X resource with the X id *xid*
            from the cache. Will raise a KeyError if there is no such
            resource in the cache.
        """
        del self._cache[xid]

    def remove_from_cache_safe(self, xid):
        """
            same as :meth:`remove_from_cache`, but does not raise a KeyError.
            Instead, it does nothing. It is safe.
        """
        try:
            self.remove_from_cache(xid)
        except KeyError:
            pass




# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ctypes

from . import (libxcb, util)
from .exception import XcbException
from .protobj import Error

class Cookie(object):
    """
        A Cookie is something like a placeholder for a request's response.
        When invoking a request method, you will get a
        :class:`ooxcb.cookie.Cookie` subclass instance as return value.
        If you need the reply, you just call the
        :meth:`ooxcb.cookie.Cookie.reply` method and you get a
        :class:`ooxcb.response.Response` subclass instance. The request
        is sent and handled just when you call `reply`.
        If you have a request without a response (a void request), you
        have to call the `..._checked` method to be able to check this
        specific cookie using the :meth:`ooxcb.cookie.Cookie.check`
        method.

        See http://lists.freedesktop.org/archives/xorg-announce/2006-September/000134.html
        for a better description of checked and unchecked requests :)
    """
    def __init__(self):
        self.conn = None
        self.reply_cls = None
        self.request = None
        self.cookie = libxcb.xcb_void_cookie_t()

    def check(self):
        """
            Explicitly check a void, checked request.
            Raises an error if there is one. If there is none,
            None is returned.
            That will raise an `XcbException` if the request is not checked
            and/or not void.
        """
        error = ctypes.pointer(libxcb.xcb_generic_error_t())
        if not (self.request.is_void and self.request.is_checked):
            raise XcbException('Request is not void and checked!')
        self.conn.check_conn()

        error = libxcb.xcb_request_check(self.conn.conn, self.cookie)
        Error.set(self.conn, error)

    def reply(self):
        """
            wait for the reply of the request and return it.
            That will raise a `XcbException` if the request is void.
        """
        if self.request.is_void:
            raise XcbException('Request has no reply.')
        self.conn.check_conn()

        err = libxcb.xcb_generic_error_t()
        error = ctypes.pointer(err)
        data = libxcb.xcb_wait_for_reply(self.conn.conn,
                self.cookie.sequence, ctypes.byref(error))
        Error.set(self.conn, error)
        if not data:
            raise IOError("I/O error on X server connection.")
        return self.reply_cls.create_from_address(self.conn, data)

    def poll_for_reply(self):
        """
            return the reply of the request if there is one. If not,
            *don't* wait for the reply to come, just return None.
            That will raise a `XcbException` if the request is void.
        """
        if self.request.is_void:
            raise XcbException('Request has no reply.')
        self.conn.check_conn()
        self.conn.flush()

        err = libxcb.xcb_generic_error_t()
        error = ctypes.pointer(err)
        data = ctypes.c_void_p()
        reply_available = libxcb.xcb_poll_for_reply(self.conn.conn,
                self.cookie.sequence, ctypes.byref(data), ctypes.byref(error))
        if reply_available:
            Error.set(self.conn, error)
            if not data:
                raise IOError("I/O error on X server connection.")
            return self.reply_cls.create_from_address(self.conn, data.value)
        else:
            return None

class VoidCookie(Cookie):
    """
        a void cookie class.
    """
    pass

__all__ = ['Cookie', 'VoidCookie']



# auto generated. yay.
import ooxcb
from ooxcb.resource import get_internal
from ooxcb.types import SIZES, make_array, build_list
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO
from struct import pack, unpack, calcsize
from ooxcb.protocol.xproto import Window
from ooxcb.protocol.xfixes import Region
from ooxcb.util import Mixin

def unpack_from_stream(fmt, stream, offset=0):
    if offset:
        stream.seek(offset, 1)
    s = stream.read(calcsize(fmt))
    return unpack(fmt, s)


MAJOR_VERSION = 0
MINOR_VERSION = 3
key = ooxcb.ExtensionKey("Composite")

class Redirect(object):
    Automatic = 0
    Manual = 1

class GetOverlayWindowReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.overlay_win = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xxxxxxxxIxxxxxxxxxxxxxxxxxxxx", stream)
        self.overlay_win = Window(self.conn, _unpacked[0])

    def build(self, stream):
        count = 0
        stream.write(pack("=xxxxxxxxIxxxxxxxxxxxxxxxxxxxx", get_internal(self.overlay_win)))

class WindowMixin(Mixin):
    target_class = Window
    def redirect_checked(self, update=Redirect.Automatic):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", window, update))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, True, True), \
            ooxcb.VoidCookie())

    def redirect(self, update=Redirect.Automatic):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", window, update))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, True, False), \
            ooxcb.VoidCookie())

    def redirect_subwindows_checked(self, update=Redirect.Automatic):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", window, update))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, True), \
            ooxcb.VoidCookie())

    def redirect_subwindows(self, update=Redirect.Automatic):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", window, update))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, False), \
            ooxcb.VoidCookie())

    def unredirect_checked(self, update=Redirect.Automatic):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", window, update))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, True), \
            ooxcb.VoidCookie())

    def unredirect(self, update=Redirect.Automatic):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", window, update))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, False), \
            ooxcb.VoidCookie())

    def unredirect_subwindows_checked(self, update=Redirect.Automatic):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", window, update))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 4, True, True), \
            ooxcb.VoidCookie())

    def unredirect_subwindows(self, update=Redirect.Automatic):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxx", window, update))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 4, True, False), \
            ooxcb.VoidCookie())

    def name_pixmap_checked(self, pixmap):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", window, pixmap))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 6, True, True), \
            ooxcb.VoidCookie())

    def name_pixmap(self, pixmap):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", window, pixmap))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 6, True, False), \
            ooxcb.VoidCookie())

    def get_overlay_window(self):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", window))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 7, False, True), \
            GetOverlayWindowCookie(),
            GetOverlayWindowReply)

    def get_overlay_window_unchecked(self):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", window))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 7, False, False), \
            GetOverlayWindowCookie(),
            GetOverlayWindowReply)

    def release_overlay_window_checked(self):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", window))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 8, True, True), \
            ooxcb.VoidCookie())

    def release_overlay_window(self):
        window = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", window))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 8, True, False), \
            ooxcb.VoidCookie())

class QueryVersionReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.major_version = None
        self.minor_version = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xxxxxxxxIIxxxxxxxxxxxxxxxx", stream)
        self.major_version = _unpacked[0]
        self.minor_version = _unpacked[1]

    def build(self, stream):
        count = 0
        stream.write(pack("=xxxxxxxxIIxxxxxxxxxxxxxxxx", self.major_version, self.minor_version))

class QueryVersionCookie(ooxcb.Cookie):
    pass

class GetOverlayWindowCookie(ooxcb.Cookie):
    pass

class RegionMixin(Mixin):
    target_class = Region
    @classmethod
    def create_from_border_clip(cls, conn, window):
        rid = conn.generate_id()
        region = cls(conn, rid)
        conn.composite.create_region_from_border_clip_checked(region, window).check()
        conn.add_to_cache(rid, region)
        return region

class compositeExtension(ooxcb.Extension):
    header = "composite"
    def query_version(self, client_major_version, client_minor_version):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", client_major_version, client_minor_version))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, True), \
            QueryVersionCookie(),
            QueryVersionReply)

    def query_version_unchecked(self, client_major_version, client_minor_version):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", client_major_version, client_minor_version))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, False), \
            QueryVersionCookie(),
            QueryVersionReply)

    def create_region_from_border_clip_checked(self, region, window):
        region = get_internal(region)
        window = get_internal(window)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", region, window))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 5, True, True), \
            ooxcb.VoidCookie())

    def create_region_from_border_clip(self, region, window):
        region = get_internal(region)
        window = get_internal(window)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", region, window))
        return self.conn.composite.send_request(ooxcb.Request(self.conn, buf.getvalue(), 5, True, False), \
            ooxcb.VoidCookie())

_events = {
}

_errors = {
}

for ev in _events.itervalues():
    if isinstance(ev.event_target_class, str):
        ev.event_target_class = globals()[ev.event_target_class]

ooxcb._add_ext(key, compositeExtension, _events, _errors)
def mixin():
    WindowMixin.mixin()
    RegionMixin.mixin()





# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ctypes
import time
import sys

from . import (libxcb, exception)
from .conn import Connection
from ooxcb.constant import *

SETUP = None

CORE = None
CORE_MODULE = None
CORE_EVENTS = {}
CORE_ERRORS = {}

EXTDICT = {}
EXT_EVENTS = {}
EXT_ERRORS = {}

def parse_auth(authstr):
    """
        Parse the X11 authentication string *authstr* and return
        a new `libxcb.xcb_auth_info` instance.
    """
    name, data = authstr.split(':')
    auth = libxcb.xcb_auth_info()
    auth.namelen = len(name)
    auth.name = name
    auth.datalen = len(data)
    auth.data = data
    return auth

def connect(display='', fd=None, auth_string=None, cls=Connection):
    """
        establishes a connection to a X display.

        The X display to connect to can be specified ...

         * by *display*, an X display string.
           If *display* is an empty string, the DISPLAY environment
           variable will be used.
         * by *fd*, an open unix file descriptor.
           *auth_string* is not relevant here.

        *cls* can be used to customize the connection class.
        There has to be an X core protocol module loaded, otherwise an
        :class:`XcbException` is raised.
        :func:`connect` returns a ready to use instance of *cls*.
    """
    auth = None

    if CORE is None:
        raise exception.XcbException(
                "No core protocol object has been set. "
                "Did you import ooxcb.protocol.xproto?"
                )

    conn = cls(CORE)
    if auth_string is not None:
        auth = parse_auth(auth_string)

    if fd is not None:
        conn.conn = libxcb.xcb_connect_to_fd(fd, ctypes.byref(auth))
    elif auth is not None:
        pref_screen = ctypes.c_int()
        conn.conn = libxcb.xcb_connect_to_display_with_auth_info(
                display, ctypes.byref(auth), ctypes.byref(pref_screen)
        )
        conn.pref_screen = pref_screen.value
    else:
        pref_screen = ctypes.c_int()
        conn.conn = libxcb.xcb_connect(display, ctypes.byref(pref_screen))
        conn.pref_screen = pref_screen.value

    conn.do_initial_setup()
    return conn

def popcount(i):
    """
        just a wrapper for the libxcb `xcb_popcount` function that determines
        the number of asserted bits in *i*.
    """
    return libxcb.xcb_popcount(i)

def type_pad(t, i):
    """
        calculates the needed type pad. Mostly for internal use.
    """
    return -i & (3 if t > 4 else t - 1)

def _add_core(value, setup, events, errors):
    """
        called by the core protocol module. Do not call it yourself
        and do not use two core modules :)
    """
    global CORE, CORE_MODULE, CORE_EVENTS, CORE_ERRORS, SETUP # eeeeeevil
    # TODO: I skipped the error checking blah blah
    if CORE is not None:
        raise XcbException("There is already a core module loaded (%s, module %r)" % (CORE, CORE_MODULE))
    CORE_MODULE = sys.modules[value.__module__]
    CORE = value
    CORE_EVENTS = events
    CORE_ERRORS = errors
    SETUP = setup

def _add_ext(key, value, events, errors):
    """
        called by X extension modules. Do not call it yourself.
    """
    EXTDICT[key] = value
    EXT_EVENTS[key] = events
    EXT_ERRORS[key] = errors

def _resize_obj(obj, size):
    """
        sets `obj.size = size`.
    """
    obj.size = size

def timestamp():
    """
        return a cardinal timestamp. a shortcut.
    """
    return int(time.time())

from ooxcb.list import *
from ooxcb.protobj import *
from ooxcb.exception import *
from ooxcb.cookie import *
from ooxcb.ext import *
from ooxcb.resource import *



# -*- coding: utf-8 -*-
# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
    This module is a direct C -> Python port of the keysyms module from
    xcb-util which can be found here: http://cgit.freedesktop.org/xcb/util

    xcb-keysyms license::

        Copyright © 2008 Ian Osgood <iano@quirkster.com>
        Copyright © 2008 Jamey Sharp <jamey@minilop.net>
        Copyright © 2008 Josh Triplett <josh@freedesktop.org>
        Copyright © 2008 Ulrich Eckhardt <doomster@knuut.de>

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without
        restriction, including without limitation the rights to use, copy,
        modify, merge, publish, distribute, sublicense, and/or sell copies
        of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
        NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
        CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
        WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

        Except as contained in this notice, the names of the authors or
        their institutions shall not be used in advertising or otherwise to
        promote the sale, use or other dealings in this Software without
        prior written authorization from the authors.

    .. data:: X_KEYS

        list of X key names
"""

from . import keysymdef
from .util import cached_property

NO_SYMBOL = 0

def convert_case(sym):
    """
        return (lower, upper) for internal use.

        :note: direct port of
               http://cgit.freedesktop.org/xcb/util/tree/keysyms/keysyms.c#n361
    """
    lower = sym
    upper = sym

    enc = sym >> 8

    if enc == 0: # latin1
        if ((sym >= keysymdef.keysyms["A"]) and (sym <= keysymdef.keysyms["Z"])):
            lower += (keysymdef.keysyms["a"] - keysymdef.keysyms["A"])
        elif ((sym >= keysymdef.keysyms["a"]) and (sym <= keysymdef.keysyms["z"])):
            upper -= (keysymdef.keysyms["a"] - keysymdef.keysyms["A"])
        elif ((sym >= keysymdef.keysyms["Agrave"])
                and (sym <= keysymdef.keysyms["Odiaeresis"])):
            lower += (keysymdef.keysyms["agrave"] - keysymdef.keysyms["Agrave"])
        elif ((sym >= keysymdef.keysyms["agrave"])
                and (sym <= keysymdef.keysyms["odiaeresis"])):
            upper -= (keysymdef.keysyms["agrave"] - keysymdef.keysyms["Agrave"])
        elif ((sym >= keysymdef.keysyms["Ooblique"]) and (sym <= keysymdef.keysyms["Thorn"])):
            lower += (keysymdef.keysyms["oslash"] - keysymdef.keysyms["Ooblique"])
        elif ((sym >= keysymdef.keysyms["oslash"]) and (sym <= keysymdef.keysyms["thorn"])):
            upper -= (keysymdef.keysyms["oslash"] - keysymdef.keysyms["Ooblique"])
    elif enc == 1: # latin2
        # Assume the KeySym is a legal value (ignore discontinuities)
        if (sym == keysymdef.keysyms["Aogonek"]):
            lower = keysymdef.keysyms["aogonek"]
        elif (sym >= keysymdef.keysyms["Lstroke"] and sym <= keysymdef.keysyms["Sacute"]):
            lower += (keysymdef.keysyms["lstroke"] - keysymdef.keysyms["Lstroke"])
        elif (sym >= keysymdef.keysyms["Scaron"] and sym <= keysymdef.keysyms["Zacute"]):
            lower += (keysymdef.keysyms["scaron"] - keysymdef.keysyms["Scaron"])
        elif (sym >= keysymdef.keysyms["Zcaron"] and sym <= keysymdef.keysyms["Zabovedot"]):
            lower += (keysymdef.keysyms["zcaron"] - keysymdef.keysyms["Zcaron"])
        elif (sym == keysymdef.keysyms["aogonek"]):
            upper = keysymdef.keysyms["Aogonek"]
        elif (sym >= keysymdef.keysyms["lstroke"] and sym <= keysymdef.keysyms["sacute"]):
            upper -= (keysymdef.keysyms["lstroke"] - keysymdef.keysyms["Lstroke"])
        elif (sym >= keysymdef.keysyms["scaron"] and sym <= keysymdef.keysyms["zacute"]):
            upper -= (keysymdef.keysyms["scaron"] - keysymdef.keysyms["Scaron"])
        elif (sym >= keysymdef.keysyms["zcaron"] and sym <= keysymdef.keysyms["zabovedot"]):
            upper -= (keysymdef.keysyms["zcaron"] - keysymdef.keysyms["Zcaron"])
        elif (sym >= keysymdef.keysyms["Racute"] and sym <= keysymdef.keysyms["Tcedilla"]):
            lower += (keysymdef.keysyms["racute"] - keysymdef.keysyms["Racute"])
        elif (sym >= keysymdef.keysyms["racute"] and sym <= keysymdef.keysyms["tcedilla"]):
            upper -= (keysymdef.keysyms["racute"] - keysymdef.keysyms["Racute"])
    elif enc == 2: # latin3
        # Assume the KeySym is a legal value (ignore discontinuities)
        if (sym >= keysymdef.keysyms["Hstroke"] and sym <= keysymdef.keysyms["Hcircumflex"]):
            lower += (keysymdef.keysyms["hstroke"] - keysymdef.keysyms["Hstroke"])
        elif (sym >= keysymdef.keysyms["Gbreve"] and sym <= keysymdef.keysyms["Jcircumflex"]):
            lower += (keysymdef.keysyms["gbreve"] - keysymdef.keysyms["Gbreve"])
        elif (sym >= keysymdef.keysyms["hstroke"] and sym <= keysymdef.keysyms["hcircumflex"]):
            upper -= (keysymdef.keysyms["hstroke"] - keysymdef.keysyms["Hstroke"])
        elif (sym >= keysymdef.keysyms["gbreve"] and sym <= keysymdef.keysyms["jcircumflex"]):
            upper -= (keysymdef.keysyms["gbreve"] - keysymdef.keysyms["Gbreve"])
        elif (sym >= keysymdef.keysyms["Cabovedot"]
                and sym <= keysymdef.keysyms["Scircumflex"]):
            lower += (keysymdef.keysyms["cabovedot"] - keysymdef.keysyms["Cabovedot"])
        elif (sym >= keysymdef.keysyms["cabovedot"]
                and sym <= keysymdef.keysyms["scircumflex"]):
            upper -= (keysymdef.keysyms["cabovedot"] - keysymdef.keysyms["Cabovedot"])
    elif enc == 3: # latin4
        # Assume the KeySym is a legal value (ignore discontinuities)
        if (sym >= keysymdef.keysyms["Rcedilla"] and sym <= keysymdef.keysyms["Tslash"]):
            lower += (keysymdef.keysyms["rcedilla"] - keysymdef.keysyms["Rcedilla"])
        elif (sym >= keysymdef.keysyms["rcedilla"] and sym <= keysymdef.keysyms["tslash"]):
            upper -= (keysymdef.keysyms["rcedilla"] - keysymdef.keysyms["Rcedilla"])
        elif (sym == keysymdef.keysyms["ENG"]):
            lower = keysymdef.keysyms["eng"]
        elif (sym == keysymdef.keysyms["eng"]):
            upper = keysymdef.keysyms["ENG"]
        elif (sym >= keysymdef.keysyms["Amacron"] and sym <= keysymdef.keysyms["Umacron"]):
            lower += (keysymdef.keysyms["amacron"] - keysymdef.keysyms["Amacron"])
        elif (sym >= keysymdef.keysyms["amacron"] and sym <= keysymdef.keysyms["umacron"]):
            upper -= (keysymdef.keysyms["amacron"] - keysymdef.keysyms["Amacron"])
    elif enc == 6: # cyrillic
        # Assume the KeySym is a legal value (ignore discontinuities)
        if (sym >= keysymdef.keysyms["Serbian_DJE"]
                and sym <= keysymdef.keysyms["Serbian_DZE"]):
            lower -= (keysymdef.keysyms["Serbian_DJE"] - keysymdef.keysyms["Serbian_dje"])
        elif (sym >= keysymdef.keysyms["Serbian_dje"]
                and sym <= keysymdef.keysyms["Serbian_dze"]):
            upper += (keysymdef.keysyms["Serbian_DJE"] - keysymdef.keysyms["Serbian_dje"])
        elif (sym >= keysymdef.keysyms["Cyrillic_YU"]
                and sym <= keysymdef.keysyms["Cyrillic_HARDSIGN"]):
            lower -= (keysymdef.keysyms["Cyrillic_YU"] - keysymdef.keysyms["Cyrillic_yu"])
        elif (sym >= keysymdef.keysyms["Cyrillic_yu"]
                and sym <= keysymdef.keysyms["Cyrillic_hardsign"]):
            upper += (keysymdef.keysyms["Cyrillic_YU"] - keysymdef.keysyms["Cyrillic_yu"])
    elif enc == 7: # greek
        if (sym >= keysymdef.keysyms["Greek_ALPHAaccent"]
                and sym <= keysymdef.keysyms["Greek_OMEGAaccent"]):
            lower += (keysymdef.keysyms["Greek_alphaaccent"] -
                    keysymdef.keysyms["Greek_ALPHAaccent"])
        elif (sym >= keysymdef.keysyms["Greek_alphaaccent"]
                and sym <= keysymdef.keysyms["Greek_omegaaccent"] and
            sym != keysymdef.keysyms["Greek_iotaaccentdieresis"] and
            sym != keysymdef.keysyms["Greek_upsilonaccentdieresis"]):
            upper -= (keysymdef.keysyms["Greek_alphaaccent"] -
                    keysymdef.keysyms["Greek_ALPHAaccent"])
        elif (sym >= keysymdef.keysyms["Greek_ALPHA"]
                and sym <= keysymdef.keysyms["Greek_OMEGA"]):
            lower += (keysymdef.keysyms["Greek_alpha"] - keysymdef.keysyms["Greek_ALPHA"])
        elif (sym >= keysymdef.keysyms["Greek_alpha"]
                and sym <= keysymdef.keysyms["Greek_omega"] and
                sym != keysymdef.keysyms["Greek_finalsmallsigma"]):
            upper -= (keysymdef.keysyms["Greek_alpha"] - keysymdef.keysyms["Greek_ALPHA"])
    elif enc == 0x14: # armenian
        if (sym >= keysymdef.keysyms["Armenian_AYB"]
                and sym <= keysymdef.keysyms["Armenian_fe"]):
            lower = sym | 1
            upper = sym & ~1
    return lower, upper

class Keysyms(object):
    """
        a simple helper for keycodes and keysyms.
    """
    def __init__(self, conn):
        """
            :type conn: :class:`ooxcb.conn.Connection`
        """
        self.conn = conn

    @cached_property
    def _cookie(self):
        min_keycode = self.conn.get_setup().min_keycode
        max_keycode = self.conn.get_setup().max_keycode
        return self.conn.xproto.get_keyboard_mapping(
                min_keycode,
                max_keycode - min_keycode + 1)

    @cached_property
    def _reply(self):
        return self._cookie.reply()

    def get_keycode(self, keysym):
        """
            return the corresponding keycode for *keysym* or None.
        """
        for j in xrange(self._reply.keysyms_per_keycode):
            for keycode in xrange(self.conn.get_setup().min_keycode,
                    self.conn.get_setup().max_keycode + 1):

                if self.get_keysym(keycode, j) == keysym:
                    return keycode
        return None

    def get_keysym(self, keycode, col):
        """
            return the corresponding keysym for *keycode* in column
            *col*.

            :todo: no error checking for now :)
        """
        keysyms = self._reply.keysyms
        min_keycode = self.conn.get_setup().min_keycode
        max_keycode = self.conn.get_setup().max_keycode
        per = self._reply.keysyms_per_keycode

        #ptr = (keycode - min_keycode) * per
        keysyms = keysyms[(keycode - min_keycode) * per:]
        # TODO: error checking
        if col < 4:
            if col > 1:
                while (per > 2 and keysyms[per - 1] == NO_SYMBOL):
                    per -= 1
                if per < 3:
                    col -= 2
            if (per <= (col|1) or keysyms[col | 1] == NO_SYMBOL):
                lsym, usym = convert_case(keysyms[col & ~1])
                if not col & 1:
                    return lsym
                elif lsym == usym:
                    return 0
                else:
                    return usym
        return keysyms[col]

X_KEYS = ['Home', 'Left', 'Up', 'Right', 'Down', 'Page_Up',
          'Page_Down', 'End', 'Begin', 'BackSpace',
          'Return', 'Escape', 'KP_Enter'] + \
         ['F%d' % i for i in range(1, 36)]

def keysym_to_str(keysym):
    """
        convert a keysym to its equivalent character or
        key description and return it.
        Returns an empty for an unknown keysym.
        That's just a shortcut for :mod:`ooxcb.keysymdef`.
    """
    return keysymdef.names.get(keysym, '')

class ConversionError(Exception):
    pass

def keysym_to_char(keysym):
    """
        try to convert *keysym* (an `int`) to a character and return it as
        an unicode string.
        If it couldn't be converted or *keysym* is NoSymbol / VoidSymbol,
        a :class:`ConversionError` is raised.
        The approach is described in `http://www.cl.cam.ac.uk/~mgk25/ucs/X11.keysyms.pdf`.
        It is able to convert latin-1, unicode and legacy keysyms. Special,
        function and vendor keysyms will raise a :class:`ConversionError`.
    """
    # special keysyms
    if keysym in (0, 0x00ffffff):
        raise ConversionError("%d is a special keysym" % keysym)
    # latin-1 keysyms
    elif (0x0020 <= keysym <= 0x007e or 0x00a0 <= keysym <= 0x00ff):
        return unichr(keysym)
    # unicode keysyms
    elif (0x01000100 <= keysym <= 0x0110ffff):
        return unichr(keysym - 0x01000000)
    # legacy keysyms
    elif keysym in keysymdef.legacy_keysyms:
        return unichr(keysymdef.legacy_keysyms[keysym])
    # dunno!
    else:
        raise ConversionError("Unsupported keysym category or legacy keysym: %d" % keysym)




# Copyright (c) 2008-2011, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .protocol import xproto

def _resource(type):
    def _converter(conn, values):
        return [conn.get_from_cache_fallback(value, type) for value in values]
    return _converter

AUTO_TYPES = {
    'WINDOW': _resource(xproto.Window),
    'PIXMAP': _resource(xproto.Pixmap),
    'DRAWABLE': _resource(xproto.Drawable),
    'STRING': lambda conn, values: ''.join(map(chr, values)),
    'UTF8_STRING': lambda conn, values: ''.join(map(chr, values)).decode('utf-8'),
    'CARDINAL': lambda conn, values: values,
    'ATOM': lambda conn, values: map(conn.atoms.get_by_id, values),
    'INTEGER': lambda conn, values: values,
}

class AutotypesError(Exception):
    pass

def autoconvert_value(conn, type, values):
    """
        Return a converted version of *values*.
        :Parameters:
            `conn`
            `type`: Atom
            `values`:
                A list of numeric values descsribing the value.
    """
    type_name = type.get_name().reply().name
    if type_name not in AUTO_TYPES:
        raise AutotypesError("Don't know how to convert %r" % type_name)
    return AUTO_TYPES[type_name](conn, values)



# auto generated. yay.
import ooxcb
from ooxcb.resource import get_internal
from ooxcb.types import SIZES, make_array, build_list
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO
from struct import pack, unpack, calcsize
from ooxcb.protocol.xproto import Drawable, Rectangle
from ooxcb.util import Mixin

def unpack_from_stream(fmt, stream, offset=0):
    if offset:
        stream.seek(offset, 1)
    s = stream.read(calcsize(fmt))
    return unpack(fmt, s)


MAJOR_VERSION = 1
MINOR_VERSION = 1
key = ooxcb.ExtensionKey("DAMAGE")

class ReportLevel(object):
    RawRectangles = 0
    DeltaRectangles = 1
    BoundingBox = 2
    NonEmpty = 3

class damageExtension(ooxcb.Extension):
    header = "damage"
    def query_version(self, client_major_version, client_minor_version):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", client_major_version, client_minor_version))
        return self.conn.damage.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, True), \
            QueryVersionCookie(),
            QueryVersionReply)

    def query_version_unchecked(self, client_major_version, client_minor_version):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", client_major_version, client_minor_version))
        return self.conn.damage.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, False), \
            QueryVersionCookie(),
            QueryVersionReply)

    def create_checked(self, damage, drawable, level):
        damage = get_internal(damage)
        drawable = get_internal(drawable)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIIBxxx", damage, drawable, level))
        return self.conn.damage.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, True, True), \
            ooxcb.VoidCookie())

    def create(self, damage, drawable, level):
        damage = get_internal(damage)
        drawable = get_internal(drawable)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIIBxxx", damage, drawable, level))
        return self.conn.damage.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, True, False), \
            ooxcb.VoidCookie())

class DrawableMixin(Mixin):
    target_class = Drawable
    def damage_add_checked(self, region):
        drawable = get_internal(self)
        region = get_internal(region)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", drawable, region))
        return self.conn.damage.send_request(ooxcb.Request(self.conn, buf.getvalue(), 4, True, True), \
            ooxcb.VoidCookie())

    def damage_add(self, region):
        drawable = get_internal(self)
        region = get_internal(region)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", drawable, region))
        return self.conn.damage.send_request(ooxcb.Request(self.conn, buf.getvalue(), 4, True, False), \
            ooxcb.VoidCookie())

class DamageNotifyEvent(ooxcb.Event):
    event_name = "on_damage_notify"
    opcode = 0
    event_target_class = "Drawable"
    def __init__(self, conn):
        ooxcb.Event.__init__(self, conn)
        self.response_type = 0
        self.level = None
        self.drawable = None
        self.damage = None
        self.timestamp = None
        self.area = None
        self.geometry = None

    def read(self, stream):
        self._address = stream.address
        root = stream.tell()
        _unpacked = unpack_from_stream("=BBxxIII", stream)
        self.response_type = _unpacked[0]
        self.level = _unpacked[1]
        self.drawable = self.conn.get_from_cache_fallback(_unpacked[2], DrawableMixin)
        self.damage = self.conn.get_from_cache_fallback(_unpacked[3], Damage)
        self.timestamp = _unpacked[4]
        self.area = Rectangle.create_from_stream(self.conn, stream)
        stream.seek(ooxcb.type_pad(8, stream.tell() - root), 1)
        self.geometry = Rectangle.create_from_stream(self.conn, stream)
        self.event_target = self.drawable

    def build(self, stream):
        count = 0
        root = stream.tell()
        stream.write(pack("=BBxxIII", self.response_type, self.level, get_internal(self.drawable), get_internal(self.damage), self.timestamp))
        count += 16
        self.area.build(stream)
        self.geometry.build(stream)
        stream.write("\0" * (32 - (stream.tell() - root)))

class Damage(ooxcb.Resource):
    def __init__(self, conn, xid):
        ooxcb.Resource.__init__(self, conn, xid)

    def destroy_checked(self):
        damage = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", damage))
        return self.conn.damage.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, True), \
            ooxcb.VoidCookie())

    def destroy(self):
        damage = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", damage))
        return self.conn.damage.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, False), \
            ooxcb.VoidCookie())

    def subtract_checked(self, repair, parts):
        damage = get_internal(self)
        repair = get_internal(repair)
        parts = get_internal(parts)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIII", damage, repair, parts))
        return self.conn.damage.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, True), \
            ooxcb.VoidCookie())

    def subtract(self, repair, parts):
        damage = get_internal(self)
        repair = get_internal(repair)
        parts = get_internal(parts)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIII", damage, repair, parts))
        return self.conn.damage.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, False), \
            ooxcb.VoidCookie())

    @classmethod
    def create(cls, conn, drawable, level):
        did = conn.generate_id()
        damage = cls(conn, did)
        conn.damage.create_checked(damage, drawable, level).check()
        conn.add_to_cache(did, damage)
        return damage

class BadDamage(ooxcb.ProtocolException):
    pass

class QueryVersionCookie(ooxcb.Cookie):
    pass

class DamageError(ooxcb.Error):
    def __init__(self, conn):
        ooxcb.Error.__init__(self, conn)

    def read(self, stream):
        self._address = stream.address

    def build(self, stream):
        count = 0

class QueryVersionReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.major_version = None
        self.minor_version = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xxxxxxxxIIxxxxxxxxxxxxxxxx", stream)
        self.major_version = _unpacked[0]
        self.minor_version = _unpacked[1]

    def build(self, stream):
        count = 0
        stream.write(pack("=xxxxxxxxIIxxxxxxxxxxxxxxxx", self.major_version, self.minor_version))

_events = {
    0: DamageNotifyEvent,
}

_errors = {
    0: (DamageError, BadDamage),
}

for ev in _events.itervalues():
    if isinstance(ev.event_target_class, str):
        ev.event_target_class = globals()[ev.event_target_class]

ooxcb._add_ext(key, damageExtension, _events, _errors)
def mixin():
    DrawableMixin.mixin()





# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import struct

from .protobj import Protobj

SIZES = {
        8: 'B',
        16: 'H',
        32: 'I',
        }

def make_array(data, typecode):
    """
        return a packed representation of the data *data*.

        :Parameters:
            `data`: list or str or unicode
                should be a list of numeric values, each item
                suitable for the given typecode.
                If it's a str or unicode instance, each char
                is converted to its ordinal value and then packed.
                Any element's `get_internal` method will be called
                if present.
            `typecode`: str
                one of:
                 * 'b', 8 bit signed
                 * 'B', 8 bit unsigned
                 * 'h', 16 bit signed,
                 * 'H', 16 bit unsigned
                 * 'i', 32 bit signed
                 * 'I', 32 bit unsigned
                 * 'f', float
                 * 'd', double
        :returns: a string
    """
    if isinstance(data, basestring):
        data = map(ord, data)
    for idx, item in enumerate(data):
        if hasattr(item, 'get_internal'):
            # Hell Yeah, so evil.
            data[idx] = item.get_internal()
    fmt = '=' + (typecode * len(data))
    return struct.pack(fmt, *data)

def make_void_array(data, format):
    """
        Return a packed representation of the data *data*.
        The only difference to `make_array` is that you pass the
        count of bytes per value to `make_void_array`, and all
        values are treated as unsigned.

        :param format: the count of bits to pack per value,
                       one of 8, 16, 32.
    """
    typecode = SIZES[format]
    return make_array(data, typecode)

def build_list(conn, stream, list_, type):
    """
        writes a *list_* of objects of the elementar
        data type *type* (where *type* is a :mod:`struct`-compatible
        type char) to *stream*.

        For each item, this checks if the item is a
        :class:`ooxcb.protobj.Protobj`. If it is, the object's `build` method
        is called.
    """
    for item in list_:
        if getattr(type, 'create_lazy', None):
            item = type.create_lazy(conn, item)
        if isinstance(item, Protobj):
            item.build(stream)
        else:
            stream.write(struct.pack(type, item))




# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ctypes

from . import libxcb

class Extension(object):
    """
        A wrapper for an X11 extension.
        This class provides with a :meth:`Extension.send_request`
        method. You will most likely do not have to use this
        class directly.
    """
    def __init__(self, conn, key=None):
        """
            :type conn: :class:`ooxcb.conn.Connection`
            :param key: The corresponding :class:`ooxcb.extkey.ExtensionKey`
                        instance (optional)
        """
        self.key = key
        self.conn = conn
        self.major_opcode = 0
        self.first_event = 0
        self.first_error = 0

    def send_request(self, request, cookie, reply_cls=None):
        """
            sends *request* to the X server. Then it provides
            *cookie* with the needed attributes and returns it.
            *reply_cls* is the optional reply class which will
            also be redirected to *cookie*.
            If :attr:`ooxcb.conn.Connection.synchronous_check`
            is True, this will also force every request to be
            checked and check it immediately after sending.
        """
        # TODO: remove that...
        if (self.conn.synchronous_check and request.is_void):
            request.is_checked = True

        xcb_req = libxcb.xcb_protocol_request_t()
        xcb_req.count = 2
        xcb_req.ext = (ctypes.pointer(self.key.key)
                if self.key is not None else None) # TODO?
        xcb_req.opcode = request.opcode
        xcb_req.isvoid = request.is_void

        s = request.buffer
        data = ctypes.cast(
                ctypes.create_string_buffer(s, len(s)),
                ctypes.c_void_p
        )

        xcb_parts = (libxcb.iovec * 2)()
        addr = ctypes.cast(xcb_parts, ctypes.c_void_p).value

        xcb_parts[0].iov_base = data
        xcb_parts[0].iov_len = request.size
        xcb_parts[1].iov_base = 0
        xcb_parts[1].iov_len = -xcb_parts[0].iov_len & 3 # ... O_o

        flags = libxcb.XCB_REQUEST_CHECKED if request.is_checked else 0
        seq = libxcb.xcb_send_request(self.conn.conn, flags,
                xcb_parts,
                ctypes.byref(xcb_req))
        cookie.conn = self.conn
        cookie.request = request
        cookie.reply_cls = reply_cls
        cookie.cookie.sequence = seq

        # TODO: remove that...
        if (self.conn.synchronous_check and request.is_void):
            cookie.check()
        return cookie

class ExtensionKey(object):
    """
        just a wrapper class for :class:`libxcb.xcb_extension_t`.
        :todo: explain what an extension key is
    """
    def __init__(self, name):
        self.key = libxcb.xcb_extension_t()
        self.key.name = name
        self.name = name

    def __hash__(self):
        return hash(self.name)




# Copyright (c) 2008-2011, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import division

US_POSITION = 1 << 0
US_SIZE = 1 << 1
P_POSITION = 1 << 2
P_SIZE = 1 << 3
P_MINSIZE = 1 << 4
P_MAXSIZE = 1 << 5
P_RESIZEINC = 1 << 6
P_ASPECT = 1 << 7
P_BASESIZE = 1 << 8
P_WINGRAVITY = 1 << 9

class SizeHints(object):
    """
        :class:`SizeHints` is a convenience class for parsing the
        `WM_NORMAL_HINTS` property (as defined in the icccm)

        .. data:: properties

            The tuple of allowed property values.
            
            :: 

                ('flags', 'x', 'y', 'width', 'height', 'min_width', 
                  'min_height', 'max_width', 'max_height', 'width_inc',
                  'height_inc', 'min_aspect_num', 'min_aspect_den',
                  'max_aspect_num', 'max_aspect_den', 'base_width', 
                  'base_height', 'win_gravity')

    """
    properties = ('flags', 'x', 'y', 'width', 'height', 'min_width',
              'min_height', 'max_width', 'max_height', 'width_inc',
              'height_inc', 'min_aspect_num', 'min_aspect_den',
              'max_aspect_num', 'max_aspect_den', 'base_width', 'base_height',
              'win_gravity')
 
    def __init__(self, **kwargs):
        """
            :param \*\*kwargs: Provided information, keys from 
                               :data:`SizeHints.properties`
        """
        flags = kwargs['flags']
        valid = False

        try:
            if flags & P_SIZE:
                self.base_width = kwargs['base_width']
                self.base_height = kwargs['base_height']
                valid = True
            elif flags & P_MINSIZE:
                self.base_width = kwargs['min_width']
                self.base_height = kwargs['min_height']
                valid = True
            else:
                self.base_width, self.base_height = 0, 0
            if flags & P_RESIZEINC:
                self.width_inc = kwargs['width_inc']
                self.height_inc = kwargs['height_inc']
                valid = True
            else:
                self.width_inc, self.height_inc = 0, 0
            if flags & P_MAXSIZE:
                self.max_width = kwargs['max_width']
                self.max_height = kwargs['height_inc']
                valid = True
            else:
                self.max_width, self.max_height = 0, 0
            if flags & P_MINSIZE:
                self.min_width = kwargs['min_width']
                self.min_height = kwargs['min_height']
                valid = True
            elif flags & P_BASESIZE:
                self.min_width = kwargs['base_width']
                self.min_height = kwargs['base_height']
                valid = True
            else:
                self.min_width, self.min_height = 0, 0
            if flags & P_ASPECT:
                self.min_aspect_num = kwargs['min_aspect_num']
                self.min_aspect_den = kwargs['min_aspect_den']
                self.max_aspect_num = kwargs['max_aspect_num']
                self.max_aspect_den = kwargs['max_aspect_den']
                valid = True
            else:
                self.min_aspect_num = 0
                self.min_aspect_den = 0
                self.max_aspect_num = 0
                self.max_aspect_den = 0
        except KeyError:
            valid = False

        self.valid = valid

    @classmethod
    def from_values(cls, values):
        """
            create a SizeHints instance from a list of integers,
            e.g. from a GetPropertyReply
        """
        return cls(**dict(zip(cls.properties, values)))

    def compute(self, geom):
        """
            compute *geom* in-place. If *self* is not valid,
            nothing is changed.
        """
        if not self.valid:
            # don't compute if not valid
            return

        if self.min_aspect_den > 0 and self.max_aspect_den > 0 and \
                (geom.width - self.base_width) > 0 and \
                (geom.height - self.base_height) > 0:
            dx = geom.width - self.base_width
            dy = geom.height - self.base_height
            min_ = self.min_aspect_num / self.min_aspect_den # float division
            max_ = self.max_aspect_num / self.max_aspect_den

            ratio = dx / dy
            if max_ > 0 and min_ > 0 and ratio > 0:
                if ratio < min:
                    dy = (dx * min_ + dy) / (min_ * min_ + 1)
                    dx = dy * min_
                    geom.width = int(dx + self.base_width)
                    geom.height = int(dy + self.base_height)
                elif ratio > max:
                    dy = (dx * min_ + dy) / (max_ * max_ + 1)
                    dx = dy * min_
                    geom.width = int(dx + self.base_width)
                    geom.height = int(dy + self.base_height)
        if self.min_width:
            geom.width = max(self.min_width, geom.width)
        if self.min_height:
            geom.height = max(self.min_height, geom.height)
        if self.max_width:
            geom.width = min(self.max_width, geom.width)
        if self.max_height:
            geom.height = min(self.max_height, geom.height)

        if self.width_inc:
            geom.width -= (geom.width - self.base_width) % self.width_inc
        if self.height_inc:
            geom.height -= (geom.height - self.base_height) % self.height_inc




# auto generated. yay.
import ooxcb
from ooxcb.resource import get_internal
from ooxcb.types import SIZES, make_array, build_list
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO
from struct import pack, unpack, calcsize

def unpack_from_stream(fmt, stream, offset=0):
    if offset:
        stream.seek(offset, 1)
    s = stream.read(calcsize(fmt))
    return unpack(fmt, s)


MAJOR_VERSION = 1
MINOR_VERSION = 13
key = ooxcb.ExtensionKey("RECORD")

class HType(object):
    FromServerTime = 1
    FromClientTime = 2
    FromClientSequence = 4

class CS(object):
    CurrentClients = 1
    FutureClients = 2
    AllClients = 3

class Category(object):
    FromServer = 0
    FromClient = 1
    ClientStarted = 2
    ClientDied = 3
    StartOfData = 4
    EndOfData = 5

class recordExtension(ooxcb.Extension):
    header = "record"
    def query_version(self, major_version, minor_version):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxHH", major_version, minor_version))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, True), \
            QueryVersionCookie(),
            QueryVersionReply)

    def query_version_unchecked(self, major_version, minor_version):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxHH", major_version, minor_version))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, False), \
            QueryVersionCookie(),
            QueryVersionReply)

    def create_context_checked(self, context, element_header, client_specs, ranges):
        num_client_specs = len(client_specs)
        num_ranges = len(ranges)
        context = get_internal(context)
        element_header = get_internal(element_header)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxxII", context, element_header, num_client_specs, num_ranges))
        buf.write(make_array(client_specs, "I"))
        for elt in ranges:
            elt.build(buf)
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, True, True), \
            ooxcb.VoidCookie())

    def create_context(self, context, element_header, client_specs, ranges):
        num_client_specs = len(client_specs)
        num_ranges = len(ranges)
        context = get_internal(context)
        element_header = get_internal(element_header)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxxII", context, element_header, num_client_specs, num_ranges))
        buf.write(make_array(client_specs, "I"))
        for elt in ranges:
            elt.build(buf)
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, True, False), \
            ooxcb.VoidCookie())

class ElementHeader(ooxcb.Resource):
    def __init__(self, conn, xid):
        ooxcb.Resource.__init__(self, conn, xid)

class ClientInfo(ooxcb.Struct):
    def __init__(self, conn):
        ooxcb.Struct.__init__(self, conn)
        self.client_resource = None
        self.num_ranges = None
        self.ranges = []

    def read(self, stream):
        self._address = stream.address
        root = stream.tell()
        _unpacked = unpack_from_stream("=II", stream)
        self.client_resource = Clientspec(self.conn, _unpacked[0])
        self.num_ranges = _unpacked[1]
        self.ranges = ooxcb.List(self.conn, stream, self.num_ranges, Range, 24)
        self.size = stream.tell() - root

    def build(self, stream):
        count = 0
        stream.write(pack("=II", get_internal(self.client_resource), self.num_ranges))
        count += 8
        build_list(self.conn, stream, self.ranges, Range)

class BadContext(ooxcb.ProtocolException):
    pass

class ContextError(ooxcb.Error):
    def __init__(self, conn):
        ooxcb.Error.__init__(self, conn)
        self.invalid_record = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xxxxI", stream)
        self.invalid_record = _unpacked[0]

    def build(self, stream):
        count = 0
        stream.write(pack("=xxxxI", self.invalid_record))

class EnableContextCookie(ooxcb.Cookie):
    pass

class QueryVersionCookie(ooxcb.Cookie):
    pass

class GetContextCookie(ooxcb.Cookie):
    pass

class GetContextReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.enabled = None
        self.element_header = None
        self.num_intercepted_clients = None
        self.intercepted_clients = []

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xBxxxxxxBxxxIxxxxxxxxxxxxxxxx", stream)
        self.enabled = _unpacked[0]
        self.element_header = ElementHeader(self.conn, _unpacked[1])
        self.num_intercepted_clients = _unpacked[2]
        self.intercepted_clients = ooxcb.List(self.conn, stream, self.num_intercepted_clients, ClientInfo, -1)

    def build(self, stream):
        count = 0
        stream.write(pack("=xBxxxxxxBxxxIxxxxxxxxxxxxxxxx", self.enabled, get_internal(self.element_header), self.num_intercepted_clients))
        count += 32
        build_list(self.conn, stream, self.intercepted_clients, ClientInfo)

class Range16(ooxcb.Struct):
    def __init__(self, conn):
        ooxcb.Struct.__init__(self, conn)
        self.first = None
        self.last = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=HH", stream)
        self.first = _unpacked[0]
        self.last = _unpacked[1]

    def build(self, stream):
        count = 0
        stream.write(pack("=HH", self.first, self.last))

    @classmethod
    def create(cls, conn, first, last):
        self = cls(conn)
        self.first = first
        self.last = last
        return self

class ExtRange(ooxcb.Struct):
    def __init__(self, conn):
        ooxcb.Struct.__init__(self, conn)
        self.major = None
        self.minor = None

    def read(self, stream):
        self._address = stream.address
        root = stream.tell()
        self.major = Range8.create_from_stream(self.conn, stream)
        stream.seek(ooxcb.type_pad(4, stream.tell() - root), 1)
        self.minor = Range16.create_from_stream(self.conn, stream)

    def build(self, stream):
        count = 0
        self.major.build(stream)
        self.minor.build(stream)

    @classmethod
    def create(cls, conn, major_first, major_last, minor_first, minor_last):
        self = cls(conn)
        self.major = Range8.create(conn, major_first, major_last)
        self.minor = Range16.create(conn, minor_first, minor_last)
        return self

class Range(ooxcb.Struct):
    def __init__(self, conn):
        ooxcb.Struct.__init__(self, conn)
        self.core_requests = None
        self.core_replies = None
        self.ext_requests = None
        self.ext_replies = None
        self.delivered_events = None
        self.device_events = None
        self.errors = None
        self.client_started = None
        self.client_died = None

    def read(self, stream):
        self._address = stream.address
        root = stream.tell()
        self.core_requests = Range8.create_from_stream(self.conn, stream)
        stream.seek(ooxcb.type_pad(2, stream.tell() - root), 1)
        self.core_replies = Range8.create_from_stream(self.conn, stream)
        stream.seek(ooxcb.type_pad(6, stream.tell() - root), 1)
        self.ext_requests = ExtRange.create_from_stream(self.conn, stream)
        stream.seek(ooxcb.type_pad(6, stream.tell() - root), 1)
        self.ext_replies = ExtRange.create_from_stream(self.conn, stream)
        stream.seek(ooxcb.type_pad(2, stream.tell() - root), 1)
        self.delivered_events = Range8.create_from_stream(self.conn, stream)
        stream.seek(ooxcb.type_pad(2, stream.tell() - root), 1)
        self.device_events = Range8.create_from_stream(self.conn, stream)
        stream.seek(ooxcb.type_pad(2, stream.tell() - root), 1)
        self.errors = Range8.create_from_stream(self.conn, stream)
        stream.seek(ooxcb.type_pad(4, stream.tell() - root), 1)
        _unpacked = unpack_from_stream("=BB", stream)
        self.client_started = _unpacked[0]
        self.client_died = _unpacked[1]

    def build(self, stream):
        count = 0
        self.core_requests.build(stream)
        self.core_replies.build(stream)
        self.ext_requests.build(stream)
        self.ext_replies.build(stream)
        self.delivered_events.build(stream)
        self.device_events.build(stream)
        self.errors.build(stream)
        stream.write(pack("=BB", self.client_started, self.client_died))

    @classmethod
    def create(cls, conn, core_requests, core_replies, ext_requests, ext_replies, delivered_events, device_events, errors, client_started, client_died):
        self = cls(conn)
        self.core_requests = Range8.create(conn, *core_requests)
        self.core_replies = Range8.create(conn, *core_replies)
        self.ext_requests = ExtRange.create(conn, *ext_requests)
        self.ext_replies = ExtRange.create(conn, *ext_replies)
        self.delivered_events = Range8.create(conn, *delivered_events)
        self.device_events = Range8.create(conn, *device_events)
        self.errors = Range8.create(conn, *errors)
        self.client_started = client_started
        self.client_died = client_died
        return self

class Context(ooxcb.Resource):
    def __init__(self, conn, xid):
        ooxcb.Resource.__init__(self, conn, xid)

    def register_clients_checked(self, element_header, client_specs, ranges):
        num_client_specs = len(client_specs)
        num_ranges = len(ranges)
        context = get_internal(self)
        element_header = get_internal(element_header)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxxII", context, element_header, num_client_specs, num_ranges))
        buf.write(make_array(client_specs, "I"))
        for elt in ranges:
            elt.build(buf)
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, True), \
            ooxcb.VoidCookie())

    def register_clients(self, element_header, client_specs, ranges):
        num_client_specs = len(client_specs)
        num_ranges = len(ranges)
        context = get_internal(self)
        element_header = get_internal(element_header)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIBxxxII", context, element_header, num_client_specs, num_ranges))
        buf.write(make_array(client_specs, "I"))
        for elt in ranges:
            elt.build(buf)
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, False), \
            ooxcb.VoidCookie())

    def unregister_clients_checked(self, client_specs):
        num_client_specs = len(client_specs)
        context = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", context, num_client_specs))
        buf.write(make_array(client_specs, "I"))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, True), \
            ooxcb.VoidCookie())

    def unregister_clients(self, client_specs):
        num_client_specs = len(client_specs)
        context = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", context, num_client_specs))
        buf.write(make_array(client_specs, "I"))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, False), \
            ooxcb.VoidCookie())

    def get(self):
        context = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", context))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 4, False, True), \
            GetContextCookie(),
            GetContextReply)

    def get_unchecked(self):
        context = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", context))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 4, False, False), \
            GetContextCookie(),
            GetContextReply)

    def enable(self):
        context = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", context))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 5, False, True), \
            EnableContextCookie(),
            EnableContextReply)

    def enable_unchecked(self):
        context = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", context))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 5, False, False), \
            EnableContextCookie(),
            EnableContextReply)

    def disable_checked(self):
        context = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", context))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 6, True, True), \
            ooxcb.VoidCookie())

    def disable(self):
        context = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", context))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 6, True, False), \
            ooxcb.VoidCookie())

    def free_checked(self):
        context = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", context))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 7, True, True), \
            ooxcb.VoidCookie())

    def free(self):
        context = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", context))
        return self.conn.record.send_request(ooxcb.Request(self.conn, buf.getvalue(), 7, True, False), \
            ooxcb.VoidCookie())

    @classmethod
    def create(cls, conn, element_header, client_specs, ranges):
        xid = conn.generate_id()
        ctx = Context(conn, xid)
        conn.record.create_context_checked(ctx, element_header, client_specs, ranges).check()
        conn.add_to_cache(xid, ctx)
        return ctx

class Clientspec(ooxcb.Resource):
    def __init__(self, conn, xid):
        ooxcb.Resource.__init__(self, conn, xid)

class Range8(ooxcb.Struct):
    def __init__(self, conn):
        ooxcb.Struct.__init__(self, conn)
        self.first = None
        self.last = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=BB", stream)
        self.first = _unpacked[0]
        self.last = _unpacked[1]

    def build(self, stream):
        count = 0
        stream.write(pack("=BB", self.first, self.last))

    @classmethod
    def create(cls, conn, first, last):
        self = cls(conn)
        self.first = first
        self.last = last
        return self

class EnableContextReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.category = None
        self.element_header = None
        self.client_swapped = None
        self.xid_base = None
        self.server_time = None
        self.rec_sequence_num = None
        self.data = []

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xBxxxxxxBBxxIIIxxxxxxxx", stream)
        self.category = _unpacked[0]
        self.element_header = ElementHeader(self.conn, _unpacked[1])
        self.client_swapped = _unpacked[2]
        self.xid_base = _unpacked[3]
        self.server_time = _unpacked[4]
        self.rec_sequence_num = _unpacked[5]
        self.data = ooxcb.List(self.conn, stream, (self.length * 4), 'B', 1)

    def build(self, stream):
        count = 0
        stream.write(pack("=xBxxxxxxBBxxIIIxxxxxxxx", self.category, get_internal(self.element_header), self.client_swapped, self.xid_base, self.server_time, self.rec_sequence_num))
        count += 32
        build_list(self.conn, stream, self.data, 'B')

class QueryVersionReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.major_version = None
        self.minor_version = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xxxxxxxxHH", stream)
        self.major_version = _unpacked[0]
        self.minor_version = _unpacked[1]

    def build(self, stream):
        count = 0
        stream.write(pack("=xxxxxxxxHH", self.major_version, self.minor_version))

_events = {
}

_errors = {
    0: (ContextError, BadContext),
}

for ev in _events.itervalues():
    if isinstance(ev.event_target_class, str):
        ev.event_target_class = globals()[ev.event_target_class]

ooxcb._add_ext(key, recordExtension, _events, _errors)
def mixin():
    pass





# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#!/usr/bin/env python

from setuptools import setup

setup(name='ooxcb',
    version='1.2',
    description='An object-oriented X Python Binding, based on xpyb',
    packages=['ooxcb', 'ooxcb.contrib', 'ooxcb.protocol'],
    keywords='xcb cairo',
    url='http://samurai-x.org',
    download_url='http://samurai-x.org/downloads/',
    author='Friedrich Weber',
    author_email='fred@reichbier.de',
)




# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# TODO: rename `ooxcb.types`, so we can do `from types import FunctionType` here
def _f():
    pass

FunctionType = type(_f)
del _f

class cached_property(object):
    """
        A simple cached property descriptor.
        from http://ronny.uberhost.de/simple-cached-for-properties-done-right
    """
    def __init__(self, func):
        self.func = func
        self.name = func.__name__

    def __get__(self, obj, type=None):
        if obj is None:
            return self

        result = self.func(obj)
        setattr(obj, self.name, result)
        return result

class cached_classproperty(object):
    """
        a modified version of :class:`cached_property` that allows to define
        cached properties on classes.
    """
    def __init__(self, func):
        self.func = func
        self.name = func.__name__

    def __get__(self, obj, type):
        # equal if we call it bound to an instance or not.
        result = self.func(type)
        setattr(type, self.name, result)
        return result

def mixin_functions(functions, into):
    """
        Add all functions in *functions* to the class *into*.
    """
    for function in functions:
        setattr(into, function.__name__, function)

def _inject_class(cls, meth):
    def classmeth(_, *args, **kwargs):
        return meth.__get__(None, cls)(*args, **kwargs)
    return classmethod(classmeth)

class MixinMeta(type):
    """
        a metaclass for mixin classes that transforms all methods
        to static methods. So the user is able to mix them into
        the class *or* to call them manually if he wants that.
    """
    def __new__(mcs, name, bases, dct):
        dct['methods'] = methods = {}
        for name in dct.keys():
            obj = dct[name]
            if isinstance(obj, FunctionType): # TODO: add staticmethod?
                methods[name] = obj
                dct[name] = staticmethod(obj)
            elif isinstance(obj, classmethod):
                if name != 'mixin': # don't wrap the builtin mixin method
                    methods[name] = obj
                    dct[name] = _inject_class(dct['target_class'], obj)
        return type.__new__(mcs, name, bases, dct)

class Mixin(object):
    """
        Base class for all mixins.
        They should have one class attribute set:

        .. attribute:: target_class

            The class this class provides some additional methods for.

        Do not try to instantiate.
    """
    __metaclass__ = MixinMeta
    target_class = None
    methods = {}

    def __init__(self, *args, **kwargs):
        raise RuntimeError("You can't instantiate mixin classes.")

    @classmethod
    def mixin(cls):
        """
            mix all additional methods into the target class.
        """
        assert cls.target_class is not None
        for name, function in cls.methods.iteritems():
            setattr(cls.target_class, name, function)




# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
    This module contains some auto-generated wrapper functions for libxcb.
    It is generated by `pyglet`_'s `wraptypes`_.

    :todo: not everything is documented yet.

    .. _pyglet: http://pyglet.org
    .. _wraptypes: http://pyglet.googlecode.com/svn/trunk/tools/wraptypes/
"""

from ctypes import *
from ctypes import util

def load_lib(name):
    libname = util.find_library(name)
    if not libname:
        raise OSError("Could not find library '%s'" % name)
    else:
        return CDLL(libname)

_lib = load_lib('xcb')

class c_void(Structure):
    # c_void_p is a buggy return type, converting to int, so
    # POINTER(None) == c_void_p is actually written as
    # POINTER(c_void), so it can be treated as a real pointer.
    _fields_ = [('dummy', c_int)]

XCB_REQUEST_CHECKED = 1 << 0
XCB_REQUEST_RAW = 1 << 1
XCB_REQUEST_DISCARD_REPLY = 1 << 2


X_PROTOCOL = 11 	# /usr/include/xcb/xcb.h:60
X_PROTOCOL_REVISION = 0 	# /usr/include/xcb/xcb.h:63
X_TCP_PORT = 6000 	# /usr/include/xcb/xcb.h:66

class struct_xcb_connection_t(Structure):
    __slots__ = [
    ]
struct_xcb_connection_t._fields_ = [
    ('_opaque_struct', c_int)
]

xcb_connection_t = struct_xcb_connection_t 	# /usr/include/xcb/xcb.h:77

class struct_anon_25(Structure):
    __slots__ = [
        'data',
        'rem',
        'index',
    ]
struct_anon_25._fields_ = [
    ('data', POINTER(None)),
    ('rem', c_int),
    ('index', c_int),
]

xcb_generic_iterator_t = struct_anon_25 	# /usr/include/xcb/xcb.h:91
class struct_anon_26(Structure):
    __slots__ = [
        'response_type',
        'pad0',
        'sequence',
        'length',
    ]
struct_anon_26._fields_ = [
    ('response_type', c_uint8),
    ('pad0', c_uint8),
    ('sequence', c_uint16),
    ('length', c_uint32),
]

xcb_generic_reply_t = struct_anon_26 	# /usr/include/xcb/xcb.h:103
class struct_anon_27(Structure):
    __slots__ = [
        'response_type',
        'pad0',
        'sequence',
        'pad',
        'full_sequence',
    ]
struct_anon_27._fields_ = [
    ('response_type', c_uint8),
    ('pad0', c_uint8),
    ('sequence', c_uint16),
    ('pad', c_uint32 * 7),
    ('full_sequence', c_uint32),
]

xcb_generic_event_t = struct_anon_27 	# /usr/include/xcb/xcb.h:116
class struct_anon_28(Structure):
    __slots__ = [
        'response_type',
        'pad0',
        'sequence',
        'length',
        'event_type',
        'pad1',
        'pad',
        'full_sequence',
    ]
struct_anon_28._fields_ = [
    ('response_type', c_uint8),
    ('pad0', c_uint8),
    ('sequence', c_uint16),
    ('length', c_uint32),
    ('event_type', c_uint16),
    ('pad1', c_uint16),
    ('pad', c_uint32 * 5),
    ('full_sequence', c_uint32),
]

xcb_ge_event_t = struct_anon_28 	# /usr/include/xcb/xcb.h:133
class struct_anon_29(Structure):
    __slots__ = [
        'response_type',
        'error_code',
        'sequence',
        'pad',
        'full_sequence',
    ]
struct_anon_29._fields_ = [
    ('response_type', c_uint8),
    ('error_code', c_uint8),
    ('sequence', c_uint16),
    ('pad', c_uint32 * 7),
    ('full_sequence', c_uint32),
]

xcb_generic_error_t = struct_anon_29 	# /usr/include/xcb/xcb.h:146
class struct_anon_30(Structure):
    __slots__ = [
        'sequence',
    ]
struct_anon_30._fields_ = [
    ('sequence', c_uint),
]

xcb_void_cookie_t = struct_anon_30 	# /usr/include/xcb/xcb.h:155
XCB_NONE = 0 	# /usr/include/xcb/xcb.h:163
XCB_COPY_FROM_PARENT = 0 	# /usr/include/xcb/xcb.h:166
XCB_CURRENT_TIME = 0 	# /usr/include/xcb/xcb.h:169
XCB_NO_SYMBOL = 0 	# /usr/include/xcb/xcb.h:172
class struct_xcb_auth_info_t(Structure):
    __slots__ = [
        'namelen',
        'name',
        'datalen',
        'data',
    ]
struct_xcb_auth_info_t._fields_ = [
    ('namelen', c_int),
    ('name', c_char_p),
    ('datalen', c_int),
    ('data', c_char_p),
]

xcb_auth_info_t = struct_xcb_auth_info_t 	# /usr/include/xcb/xcb.h:187
# /usr/include/xcb/xcb.h:200
xcb_flush = _lib.xcb_flush
xcb_flush.restype = c_int
xcb_flush.argtypes = [POINTER(xcb_connection_t)]

# /usr/include/xcb/xcb.h:217
xcb_get_maximum_request_length = _lib.xcb_get_maximum_request_length
xcb_get_maximum_request_length.restype = c_uint32
xcb_get_maximum_request_length.argtypes = [POINTER(xcb_connection_t)]

# /usr/include/xcb/xcb.h:236
xcb_prefetch_maximum_request_length = _lib.xcb_prefetch_maximum_request_length
xcb_prefetch_maximum_request_length.restype = None
xcb_prefetch_maximum_request_length.argtypes = [POINTER(xcb_connection_t)]

# /usr/include/xcb/xcb.h:250
xcb_wait_for_event = _lib.xcb_wait_for_event
xcb_wait_for_event.restype = POINTER(xcb_generic_event_t)
xcb_wait_for_event.argtypes = [POINTER(xcb_connection_t)]

# /usr/include/xcb/xcb.h:264
xcb_poll_for_event = _lib.xcb_poll_for_event
xcb_poll_for_event.restype = POINTER(xcb_generic_event_t)
xcb_poll_for_event.argtypes = [POINTER(xcb_connection_t)]

# /usr/include/xcb/xcb.h:282
xcb_request_check = _lib.xcb_request_check
xcb_request_check.restype = POINTER(xcb_generic_error_t)
xcb_request_check.argtypes = [POINTER(xcb_connection_t), xcb_void_cookie_t]

class struct_xcb_extension_t(Structure):
    _fields_ = [('name', c_char_p),
            ('global_id', c_int)] # xpyb needs it ... cannot be opaque, so changed it.

xcb_extension_t = struct_xcb_extension_t 	# /usr/include/xcb/xcb.h:290
class struct_xcb_query_extension_reply_t(Structure):
    __slots__ = [
        'response_type',
        'pad0',
        'sequence',
        'length',
        'present',
        'major_opcode',
        'first_event',
        'first_error',
    ]
struct_xcb_query_extension_reply_t._fields_ = [
    ('response_type', c_uint8),
    ('pad0', c_uint8),
    ('sequence', c_uint16),
    ('length', c_uint32),
    ('present', c_uint8),
    ('major_opcode', c_uint8),
    ('first_event', c_uint8),
    ('first_error', c_uint8),
]

xcb_query_extension_reply_t = struct_xcb_query_extension_reply_t 	# /usr/include/xcb/xproto.h:3667
# /usr/include/xcb/xcb.h:308
xcb_get_extension_data = _lib.xcb_get_extension_data
xcb_get_extension_data.restype = POINTER(xcb_query_extension_reply_t)
xcb_get_extension_data.argtypes = [POINTER(xcb_connection_t), POINTER(xcb_extension_t)]

# /usr/include/xcb/xcb.h:321
xcb_prefetch_extension_data = _lib.xcb_prefetch_extension_data
xcb_prefetch_extension_data.restype = None
xcb_prefetch_extension_data.argtypes = [POINTER(xcb_connection_t), POINTER(xcb_extension_t)]

class struct_xcb_setup_t(Structure):
    __slots__ = [
        'status',
        'pad0',
        'protocol_major_version',
        'protocol_minor_version',
        'length',
        'release_number',
        'resource_id_base',
        'resource_id_mask',
        'motion_buffer_size',
        'vendor_len',
        'maximum_request_length',
        'roots_len',
        'pixmap_formats_len',
        'image_byte_order',
        'bitmap_format_bit_order',
        'bitmap_format_scanline_unit',
        'bitmap_format_scanline_pad',
        'min_keycode',
        'max_keycode',
        'pad1',
    ]
xcb_keycode_t = c_uint8 	# /usr/include/xcb/xproto.h:166
struct_xcb_setup_t._fields_ = [
    ('status', c_uint8),
    ('pad0', c_uint8),
    ('protocol_major_version', c_uint16),
    ('protocol_minor_version', c_uint16),
    ('length', c_uint16),
    ('release_number', c_uint32),
    ('resource_id_base', c_uint32),
    ('resource_id_mask', c_uint32),
    ('motion_buffer_size', c_uint32),
    ('vendor_len', c_uint16),
    ('maximum_request_length', c_uint16),
    ('roots_len', c_uint8),
    ('pixmap_formats_len', c_uint8),
    ('image_byte_order', c_uint8),
    ('bitmap_format_bit_order', c_uint8),
    ('bitmap_format_scanline_unit', c_uint8),
    ('bitmap_format_scanline_pad', c_uint8),
    ('min_keycode', xcb_keycode_t),
    ('max_keycode', xcb_keycode_t),
    ('pad1', c_uint8 * 4),
]

xcb_setup_t = struct_xcb_setup_t 	# /usr/include/xcb/xproto.h:434
# /usr/include/xcb/xcb.h:344
xcb_get_setup = _lib.xcb_get_setup
xcb_get_setup.restype = POINTER(xcb_setup_t)
xcb_get_setup.argtypes = [POINTER(xcb_connection_t)]

# /usr/include/xcb/xcb.h:354
xcb_get_file_descriptor = _lib.xcb_get_file_descriptor
xcb_get_file_descriptor.restype = c_int
xcb_get_file_descriptor.argtypes = [POINTER(xcb_connection_t)]

# /usr/include/xcb/xcb.h:369
xcb_connection_has_error = _lib.xcb_connection_has_error
xcb_connection_has_error.restype = c_int
xcb_connection_has_error.argtypes = [POINTER(xcb_connection_t)]

# /usr/include/xcb/xcb.h:383
xcb_connect_to_fd = _lib.xcb_connect_to_fd
xcb_connect_to_fd.restype = POINTER(xcb_connection_t)
xcb_connect_to_fd.argtypes = [c_int, POINTER(xcb_auth_info_t)]

# /usr/include/xcb/xcb.h:392
xcb_disconnect = _lib.xcb_disconnect
xcb_disconnect.restype = None
xcb_disconnect.argtypes = [POINTER(xcb_connection_t)]

# /usr/include/xcb/xcb.h:414
xcb_parse_display = _lib.xcb_parse_display
xcb_parse_display.restype = c_int
xcb_parse_display.argtypes = [c_char_p, POINTER(c_char_p), POINTER(c_int), POINTER(c_int)]

# /usr/include/xcb/xcb.h:428
xcb_connect = _lib.xcb_connect
xcb_connect.restype = POINTER(xcb_connection_t)
xcb_connect.argtypes = [c_char_p, POINTER(c_int)]

# /usr/include/xcb/xcb.h:442
xcb_connect_to_display_with_auth_info = _lib.xcb_connect_to_display_with_auth_info
xcb_connect_to_display_with_auth_info.restype = POINTER(xcb_connection_t)
xcb_connect_to_display_with_auth_info.argtypes = [c_char_p, POINTER(xcb_auth_info_t), POINTER(c_int)]

# /usr/include/xcb/xcb.h:455
xcb_generate_id = _lib.xcb_generate_id
xcb_generate_id.restype = c_uint32
xcb_generate_id.argtypes = [POINTER(xcb_connection_t)]

# .... sys/uio.h

class iovec(Structure):
    _fields_ = [('iov_base', c_void_p),
            ('iov_len', c_size_t)]


# ... xcbext.h

class struct_anon_31(Structure):
    __slots__ = [
        'count',
        'ext',
        'opcode',
        'isvoid',
    ]

struct_anon_31._fields_ = [
    ('count', c_size_t),
    ('ext', POINTER(xcb_extension_t)),
    ('opcode', c_uint8),
    ('isvoid', c_uint8),
]

xcb_protocol_request_t = struct_anon_31 	# /usr/include/xcb/xcbext.h:52

# /usr/include/xcb/xcbext.h:60
xcb_send_request = _lib.xcb_send_request
xcb_send_request.restype = c_uint
xcb_send_request.argtypes = [POINTER(xcb_connection_t), c_int, POINTER(iovec), POINTER(xcb_protocol_request_t)]

class struct_anon_29(Structure):
    __slots__ = [
        'response_type',
        'error_code',
        'sequence',
        'pad',
        'full_sequence',
    ]
struct_anon_29._fields_ = [
    ('response_type', c_uint8),
    ('error_code', c_uint8),
    ('sequence', c_uint16),
    ('pad', c_uint32 * 7),
    ('full_sequence', c_uint32),
]

xcb_generic_error_t = struct_anon_29 	# /usr/include/xcb/xcb.h:146
# /usr/include/xcb/xcbext.h:65
xcb_wait_for_reply = _lib.xcb_wait_for_reply
xcb_wait_for_reply.restype = c_void_p
xcb_wait_for_reply.argtypes = [POINTER(xcb_connection_t), c_uint, POINTER(POINTER(xcb_generic_error_t))]

# /usr/include/xcb/xcbext.h:66
xcb_poll_for_reply = _lib.xcb_poll_for_reply
xcb_poll_for_reply.restype = c_int
xcb_poll_for_reply.argtypes = [POINTER(xcb_connection_t), c_uint, POINTER(POINTER(None)), POINTER(POINTER(xcb_generic_error_t))]

# /usr/include/xcb/xcbext.h:71
xcb_popcount = _lib.xcb_popcount
xcb_popcount.restype = c_int
xcb_popcount.argtypes = [c_uint32]

__all__ = [ \
# xcb.h
'X_PROTOCOL', 'X_PROTOCOL_REVISION', 'X_TCP_PORT',
'xcb_connection_t', 'xcb_generic_iterator_t', 'xcb_generic_reply_t',
'xcb_generic_event_t', 'xcb_ge_event_t', 'xcb_generic_error_t',
'xcb_void_cookie_t', 'XCB_NONE', 'XCB_COPY_FROM_PARENT', 'XCB_CURRENT_TIME',
'XCB_NO_SYMBOL', 'xcb_auth_info_t', 'xcb_flush',
'xcb_get_maximum_request_length', 'xcb_prefetch_maximum_request_length',
'xcb_wait_for_event', 'xcb_poll_for_event', 'xcb_request_check',
'xcb_extension_t', 'xcb_get_extension_data', 'xcb_prefetch_extension_data',
'xcb_get_setup', 'xcb_get_file_descriptor', 'xcb_connection_has_error',
'xcb_connect_to_fd', 'xcb_disconnect', 'xcb_parse_display', 'xcb_connect',
'xcb_connect_to_display_with_auth_info', 'xcb_generate_id',
# xcbext.h
'xcb_protocol_request_t', 'xcb_send_request', 'xcb_wait_for_reply',
'xcb_poll_for_reply', 'xcb_popcount'
# sys/uio.h
'iovec']




# auto generated. yay.
import ooxcb
from ooxcb.resource import get_internal
from ooxcb.types import SIZES, make_array, build_list
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO
from struct import pack, unpack, calcsize
from ooxcb.protocol.xproto import Drawable, Window
from ooxcb.util import Mixin

def unpack_from_stream(fmt, stream, offset=0):
    if offset:
        stream.seek(offset, 1)
    s = stream.read(calcsize(fmt))
    return unpack(fmt, s)


MAJOR_VERSION = 1
MINOR_VERSION = 1
key = ooxcb.ExtensionKey("MIT-SCREEN-SAVER")

class Kind(object):
    Blanked = 0
    Internal = 1
    External = 2

class State(object):
    Off = 0
    On = 1
    Cycle = 2
    Disabled = 3

class CW(object):
    BackPixmap = 1
    BackPixel = 2
    BorderPixmap = 4
    BorderPixel = 8
    BitGravity = 16
    WinGravity = 32
    BackingStore = 64
    BackingPlanes = 128
    BackingPixel = 256
    OverrideRedirect = 512
    SaveUnder = 1024
    EventMask = 2048
    DontPropagate = 4096
    Colormap = 8192
    Cursor = 16384

class Event(object):
    NotifyMask = 1
    CycleMask = 2

class WindowMixin(Mixin):
    target_class = Window

class QueryVersionReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.server_major_version = None
        self.server_minor_version = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xxxxxxxxHHxxxxxxxxxxxxxxxxxxxx", stream)
        self.server_major_version = _unpacked[0]
        self.server_minor_version = _unpacked[1]

    def build(self, stream):
        count = 0
        stream.write(pack("=xxxxxxxxHHxxxxxxxxxxxxxxxxxxxx", self.server_major_version, self.server_minor_version))

class screensaverExtension(ooxcb.Extension):
    header = "screensaver"
    def query_version(self, client_major_version, client_minor_version):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBBxx", client_major_version, client_minor_version))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, True), \
            QueryVersionCookie(),
            QueryVersionReply)

    def query_version_unchecked(self, client_major_version, client_minor_version):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBBxx", client_major_version, client_minor_version))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, False), \
            QueryVersionCookie(),
            QueryVersionReply)

    def suspend_checked(self, suspend):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBxxx", suspend))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 5, True, True), \
            ooxcb.VoidCookie())

    def suspend(self, suspend):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBxxx", suspend))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 5, True, False), \
            ooxcb.VoidCookie())

class DrawableMixin(Mixin):
    target_class = Drawable
    def query_info(self):
        drawable = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", drawable))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, False, True), \
            QueryInfoCookie(),
            QueryInfoReply)

    def query_info_unchecked(self):
        drawable = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", drawable))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, False, False), \
            QueryInfoCookie(),
            QueryInfoReply)

    def select_input_checked(self, event_mask):
        drawable = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", drawable, event_mask))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, True), \
            ooxcb.VoidCookie())

    def select_input(self, event_mask):
        drawable = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", drawable, event_mask))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, False), \
            ooxcb.VoidCookie())

    def set_attributes_checked(self, x, y, width, height, border_width, _class, depth, visual, value_mask, value_list):
        value_mask, value_list = 0, []
        if "back_pixmap" in values:
            value_mask |= 1
            value_list.append(get_internal(values["back_pixmap"]))
        if "back_pixel" in values:
            value_mask |= 2
            value_list.append(values["back_pixel"])
        if "border_pixmap" in values:
            value_mask |= 4
            value_list.append(values["border_pixmap"])
        if "border_pixel" in values:
            value_mask |= 8
            value_list.append(values["border_pixel"])
        if "bit_gravity" in values:
            value_mask |= 16
            value_list.append(values["bit_gravity"])
        if "win_gravity" in values:
            value_mask |= 32
            value_list.append(values["win_gravity"])
        if "backing_store" in values:
            value_mask |= 64
            value_list.append(values["backing_store"])
        if "backing_planes" in values:
            value_mask |= 128
            value_list.append(values["backing_planes"])
        if "backing_pixel" in values:
            value_mask |= 256
            value_list.append(values["backing_pixel"])
        if "override_redirect" in values:
            value_mask |= 512
            value_list.append(values["override_redirect"])
        if "save_under" in values:
            value_mask |= 1024
            value_list.append(values["save_under"])
        if "event_mask" in values:
            value_mask |= 2048
            value_list.append(values["event_mask"])
        if "dont_propagate" in values:
            value_mask |= 4096
            value_list.append(values["dont_propagate"])
        if "colormap" in values:
            value_mask |= 8192
            value_list.append(get_internal(values["colormap"]))
        if "cursor" in values:
            value_mask |= 16384
            value_list.append(get_internal(values["cursor"]))
        drawable = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIhhHHHBBII", drawable, x, y, width, height, border_width, _class, depth, visual, value_mask))
        buf.write(make_array(value_list, "I"))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, True), \
            ooxcb.VoidCookie())

    def set_attributes(self, x, y, width, height, border_width, _class, depth, visual, value_mask, value_list):
        value_mask, value_list = 0, []
        if "back_pixmap" in values:
            value_mask |= 1
            value_list.append(get_internal(values["back_pixmap"]))
        if "back_pixel" in values:
            value_mask |= 2
            value_list.append(values["back_pixel"])
        if "border_pixmap" in values:
            value_mask |= 4
            value_list.append(values["border_pixmap"])
        if "border_pixel" in values:
            value_mask |= 8
            value_list.append(values["border_pixel"])
        if "bit_gravity" in values:
            value_mask |= 16
            value_list.append(values["bit_gravity"])
        if "win_gravity" in values:
            value_mask |= 32
            value_list.append(values["win_gravity"])
        if "backing_store" in values:
            value_mask |= 64
            value_list.append(values["backing_store"])
        if "backing_planes" in values:
            value_mask |= 128
            value_list.append(values["backing_planes"])
        if "backing_pixel" in values:
            value_mask |= 256
            value_list.append(values["backing_pixel"])
        if "override_redirect" in values:
            value_mask |= 512
            value_list.append(values["override_redirect"])
        if "save_under" in values:
            value_mask |= 1024
            value_list.append(values["save_under"])
        if "event_mask" in values:
            value_mask |= 2048
            value_list.append(values["event_mask"])
        if "dont_propagate" in values:
            value_mask |= 4096
            value_list.append(values["dont_propagate"])
        if "colormap" in values:
            value_mask |= 8192
            value_list.append(get_internal(values["colormap"]))
        if "cursor" in values:
            value_mask |= 16384
            value_list.append(get_internal(values["cursor"]))
        drawable = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxIhhHHHBBII", drawable, x, y, width, height, border_width, _class, depth, visual, value_mask))
        buf.write(make_array(value_list, "I"))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, False), \
            ooxcb.VoidCookie())

    def unset_attributes_checked(self):
        drawable = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", drawable))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 4, True, True), \
            ooxcb.VoidCookie())

    def unset_attributes(self):
        drawable = get_internal(self)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxI", drawable))
        return self.conn.screensaver.send_request(ooxcb.Request(self.conn, buf.getvalue(), 4, True, False), \
            ooxcb.VoidCookie())

class NotifyEvent(ooxcb.Event):
    event_name = "on_notify"
    opcode = 0
    event_target_class = "Window"
    def __init__(self, conn):
        ooxcb.Event.__init__(self, conn)
        self.response_type = 0
        self.code = None
        self.state = None
        self.sequence_number = None
        self.time = None
        self.root = None
        self.window = None
        self.kind = None
        self.forced = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=BBxxBxHIIIBBxxxxxxxxxxxxxx", stream)
        self.response_type = _unpacked[0]
        self.code = _unpacked[1]
        self.state = _unpacked[2]
        self.sequence_number = _unpacked[3]
        self.time = _unpacked[4]
        self.root = Window(self.conn, _unpacked[5])
        self.window = Window(self.conn, _unpacked[6])
        self.kind = _unpacked[7]
        self.forced = _unpacked[8]
        self.event_target = self.window

    def build(self, stream):
        count = 0
        root = stream.tell()
        stream.write(pack("=BBxxBxHIIIBBxxxxxxxxxxxxxx", self.response_type, self.code, self.state, self.sequence_number, self.time, get_internal(self.root), get_internal(self.window), self.kind, self.forced))
        stream.write("\0" * (32 - (stream.tell() - root)))

class QueryVersionCookie(ooxcb.Cookie):
    pass

class QueryInfoCookie(ooxcb.Cookie):
    pass

class QueryInfoReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.state = None
        self.saver_window = None
        self.ms_until_server = None
        self.ms_since_user_input = None
        self.event_mask = None
        self.kind = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xBxxxxxxIIIIBxxxxxxx", stream)
        self.state = _unpacked[0]
        self.saver_window = Window(self.conn, _unpacked[1])
        self.ms_until_server = _unpacked[2]
        self.ms_since_user_input = _unpacked[3]
        self.event_mask = _unpacked[4]
        self.kind = _unpacked[5]

    def build(self, stream):
        count = 0
        stream.write(pack("=xBxxxxxxIIIIBxxxxxxx", self.state, get_internal(self.saver_window), self.ms_until_server, self.ms_since_user_input, self.event_mask, self.kind))

_events = {
    0: NotifyEvent,
}

_errors = {
}

for ev in _events.itervalues():
    if isinstance(ev.event_target_class, str):
        ev.event_target_class = globals()[ev.event_target_class]

ooxcb._add_ext(key, screensaverExtension, _events, _errors)
def mixin():
    DrawableMixin.mixin()
    WindowMixin.mixin()


